
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2008 Yann Glouche and Thomas Genet

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.3. Last modification 13/03/08
*)
open Types;;

(*
  Definition of the variable that will contain the buffer to be
  lexically analysed as specification input.
  \begin{center}
    type: [val lexbuf : Lexing.lexbuf ref]
  \end{center}
*)
let lexbuf = ref (Lexing.from_string "");;

(*

  \section{The global record of strings\label{globals:string:id}}

  The [string_id] variable records all identifiers found during 
  the parsing of the protocol. They are inserted in this list 
  using:
  \begin{center}
  [Globals.string_id#register_name s]
  \end{center}
  This function returns the id of the name in the list of strings.
  Note that type constructors try, as much as possible,
  to insert the strings into the table in a transparent way.
  A name can be retrieved from an [id] by using:
  \begin{center}
  [Globals.string_id#get_name id]
  \end{center}
  This functions returns the string associated with the identifier 
  [id].
*)

(*
  \begin{center}
   type: [val string_id : Types.ident_table]
  \end{center}
*)
let string_id = new ident_table;;

(*
  \section{Records of variables}

  Before we step into particular data structures, let us summarize
  how compilation of a protocol is handled. We take as example
  the compilation of a role [Alice], which is later called using the
  constant [a] as parameter.

  \begin{figure}
  \begin{center}
  \mygraphics{all_var_tables}
  \caption{\label{fig:tables:handling:variables}
  Tables handling the variables}
  \end{center}
  \end{figure}

  Let us comment this figure.
  \begin{enumerate}
  \item First, the translation from [string] identifiers to [int]
  identifiers is performed during the parsing of the specification.
  After this phase, variables are handled in a [hlpsl_variable] 
  structure (see Section~\ref{types:hlpsl:variable}). An object of
  the [hlpsl_variable] class is only meaningful inside a role, as
  we need a [session_ID] to transform it into a variable in the
  IF (\emph{i.e.} a variable having a [global_var_id]). 
  \item During the instantiation, the compiler
  builds the [global_var_id] table. No semantic check is performed
  while building this table, which can be seen as just an injective 
  mapping from $N\times N\to N$. At the same time, the [memory_map]
  and [initial_value] tables are build. The algorithm to create the 
  memory map is rather simple. Given variable $v$ in role $r$, with
  global identity of $\nu$,
  \begin{trivlist}
  \item\underline{\textbf{Partial construction of
  the [memory_map] table}}
  \begin{itemize}
  \item  If $v$ is in the parameter list of role $r$, consider
  the variable $t$ such that $t$ is at the position of $v$ in the call 
  of role $r$. For example, the role is declared [Role r(v:...)]
  and is called by [r(t,...)], let $x_t$ be the [global_var_id] (it exists
  by recurrence) associated with [t], and let $\alpha(x_t)$ be the value 
  of $x_t$ in the [memory_map] table. We add the field $(\nu,\alpha)$
  in the [memory_map] table. 
  \item If $v$ is declared in the [local] part of the specification,
  let $\nu$ be the [global_var_id] of $v$ in the instance, and let $n_\nu$
  be a "new" integer (we model it with a counter). We add the record
  $(\nu,n_\nu)$ to the [memory_map] table.
  \end{itemize}
  \item\underline{\textbf{Assignment of values to explicit variables}}
  \begin{itemize}
  \item For variables in the parameter list, assign the value of the
  parameter to the variable;
  \item For variables in the [local] declaration, either assign the
  value given in the [init] state or leave it as [Var(global_var_id)].
  \end{itemize}
  \item\underline{\textbf{End of assignment of values}}.
  \begin{itemize}
  \item If a term $t$ is given as parameter to a function, create a 
  new variable $x_t$ and give it a value according to the value of 
  variables in $t$;
  \item If there is a composition over a set, for exemple 
  [{t(x_1,\ldots,x_n):L}]. If $L=\{l_1,\ldots,l_m\}$, for each 
  $l_i$, create new variables $x_{1,i},\ldots,x_{n,i}$, give them a global
  variable identifier and a variable found by pattern-matching. If the 
  pattern-matching fails (because the value of $L$ contains variables,
  for example), the compilation exits with an error.
  \end{itemize}
  \end{trivlist}
  \end{enumerate}
  
  At the end of this process, all parameters are given a value, a 
  position in [memory_map] and a global identifier. It is thus possible to
  compile the roles called from this role.

  The [function_value] is used as an indirection table for functions. For 
  example, if $L$ is a function, $L(A)$ can be a public key. In this case,
  we add the variable $x_{L(A)}$ to the [memory_map] table, and give it
  a value [Func(x,y,z)], where $x$ is the global identifier of $L$, $y$
  is the global identifier of $A$, and $z$ is a key in [function_value].
  The value of $z$ in this table is the public key of $A$. The reason to
  use this trick is that, in the IF, this will permit to connect the
  public key of $A$ with $L$ and $A$. 
*)

(* First argument is the player and the second is the role
   \begin{center}
   type: [val player_to_role : (int * int) list ref]
   \end{center}
*)
let player_to_role = ref ([] : (int * int) list);; 

(*
   \begin{center}
   type: [val global_var_id : Types.global_var_ident_table]
   \end{center}
*)
let global_var_id = new global_var_ident_table;;

(*
   \begin{center}
   type: [val memory_map : Types.memory_map_table]
   \end{center}
*)
let memory_map = new memory_map_table;;

(*
   \begin{center}
   type: [val type_table : Types.type_table_type]
   \end{center}
*)
let type_table = new type_table_type;;

(*
   \begin{center}
   type: [val value_table : Types.value_map_table]
   \end{center}
*)
let value_table = new value_map_table;;


(*
  \section{Built-in values}
*)

(*
   \begin{center}
   type: [val intruder_name : string]
   type: [val global_i : int]
   \end{center}
*)
let intruder_name = "i";;
let global_i = (string_id#register_name intruder_name 0);;
ignore (type_table#register_type global_i (Base(Agent)));;


(* To register predefined constants: exp and xor *)
ignore (type_table#register_type (string_id#register_name "exp" 0) (Function(Base(Message),[Base(Message);Base(Message)])));;
ignore (type_table#register_type (string_id#register_name "xor" 0) (Function(Base(Message),[Base(Message);Base(Message)])));;


(*
   \begin{center}
   type: [val dummy_values_string_list : (Types.basic_type * string) list]
   \end{center}
*)
let dummy_values_string_list =
(*** version non abstraite : ***)
  [(Agent,         "dummy_agent");
   (Nat,           "dummy_nat");
   (Bool,          "dummy_bool");
   (Hash_func,     "dummy_hash");
   (Message,       "dummy_msg");
   (Text,          "dummy_nonce");
   (Public_key,    "dummy_pk");
   (Symmetric_key, "dummy_sk");
   (Channel,       "dummy_chnl");
   (Channel_ota,   "dummy_chnl_ota");
   (Channel_dy,   "dummy_chnl_dy");
   (Protocol_id,   "dummy_pid")
  ](***)
(*** version abstraite : ***
  [(Agent,         "Dummy_agent");
   (Nat,           "Dummy_nat");
   (Bool,          "Dummy_bool");
   (Hash_func,     "Dummy_hash");
   (Message,       "Dummy_msg");
   (Text,          "Dummy_nonce");
   (Public_key,    "Dummy_pk");
   (Symmetric_key, "Dummy_sk");
   (Channel,       "Dummy_chnl");
   (Channel_ota,   "Dummy_chnl_ota");
   (Channel_dy,   "Dummy_chnl_dy");
   (Protocol_id,   "Dummy_pid")
  ]***)
;;
(* if we want to register constants from the beginning:
List.iter (fun (t,s) ->
  ignore (type_table#register_type (string_id#register_name s 0) (Base(t))))
  dummy_values_string_list
;;*)

(*
   \begin{center}
   type: [val is_dummy_term : Types.if_value_type -> bool]
   \end{center}
*)
let rec is_dummy_term = function
    If_const(i) ->
      (match (string_id#get_name i) with
	"dummy_agent"
      | "dummy_nat"
      | "dummy_bool"
      | "dummy_hash"
      | "dummy_msg"
      | "dummy_nonce"
      | "dummy_pk"
      | "dummy_sk"
      | "dummy_chnl"
      | "dummy_chnl_ota"
      | "dummy_chnl_dy"
      | "dummy_pid"
      | "dummy_func"
      | "dummy_set"
      | "dummy_list" -> true
      | _ -> false)
  | If_pair(t1,t2)
  | If_crypt(t1,t2)
  | If_scrypt(t1,t2) -> (is_dummy_term t1) || (is_dummy_term t2)
  | If_inv(t) -> is_dummy_term t
  | _ -> false
;;


(*
   \Begin{center}
   type: [val global_start : int]
   \end{center}
let global_start = (string_id#register_name "start" 0);;
ignore (type_table#register_type global_start (Base Message));;
*)
let global_start = ref (-1);;
let register_start () =
  if !global_start = -1 then
    (global_start := (string_id#register_name "start" 0);
     ignore (type_table#register_type !global_start (Base Message)))
;;

(* global_start_event: indicates if a start constant have been encountered *)
let global_start_const = ref false;;

(* global_start_event: indicates if a start() event have been encountered *
let global_start_event = ref false;;*)


(*
   \Begin{center}
   type: [val not_excluded_type : term_type -> bool]
   \end{center}
*)
let rec not_excluded_type =  function
    (* excluded types *)
    Base(Channel)
  | Base(Channel_ota)
  | Base(Channel_dy)
  | Function(_) -> false
    (* recursive checks *)
  | Hash(t)
  | Inv(t) ->
      (not_excluded_type t)
  | Pair(t1,t2)
  | Scrypt(t1,t2)
  | Crypt(t1,t2)
  | Cons(t1,t2)
  | Delete(t1,t2) ->
      (not_excluded_type t1) && (not_excluded_type t2)
(*  | Lst(lt) *)
  | Set(lt) ->
      List.for_all not_excluded_type lt
    (* non-recursive useful types *)
  | Base(_)
  | Prime(_) -> true
;;

(* This function looks for a type declaration of a given variable,
   used in the goal section (and so not declared there)  *)
let get_any_type_of i =
  let si = string_id#get_name i
  and k = ref (-1)
  and tk = ref (Base(Nat)) in
  Hashtbl.iter
    (fun j sj ->
      try 
	if (i <> j) && (si = sj) then
	  (tk := type_table#get_type j;
	   k := j)
      with Not_found -> ())
    string_id#get_tab;
  if !k <> -1 then
    !tk
  else
    raise Not_found
;;


(*
  \section{Flags}
*)

(* We define flags that will be set up according to the command line arguments: *)
(*
   \begin{center}
   type: [val flag_types : bool ref]\\
   type: [val flag_init : bool ref]\\
   type: [val flag_rules : bool ref]\\
   type: [val flag_goals : bool ref]\\
   type: [val flag_split : bool ref]\\
   type: [val flag_stdout : bool ref]
   \end{center}
*)
let flag_types    = ref false;;
let flag_init     = ref false;;
let flag_rules    = ref false;;
let flag_goals    = ref false;;
let flag_split    = ref false;;
let flag_stdout   = ref false;;
let flag_warnings = ref true;;



(*
  \section{Output information}
   By default, the IF specification is written in the file whose name
   is in the variable [output_file].  But this printing is first done
   in a temporary file [tmp_file], located in the standard temporary
   directory.
   \begin{center}
   type: [val output_file : string ref]\\
   type: [val tmp_file : string ref]\\
   type: [val tmp_channel : out_channel ref]
   \end{center}
*)
let input_file = ref "";;
let output_dir = ref "";;
let output_file = ref "";;
let tmp_file = ref "";;
let tmp_channel = ref stdout;;

(*
   The next function exists in OCaml, but only from version 3.07.
   So it will have to be removed when everybody will use this version...
   \begin{center}
   type: [val close_out_noerr : out_channel -> unit]
   \end{center}
*)
let close_out_noerr c =
  try close_out c
  with _ -> ()
;;

(*
   The following functions are used for printing in the chosen output.
*)

(*
   \begin{center}
   type: [val prout_endline : string -> unit]
   \end{center}
*)
let prout_endline s =
  Printf.fprintf !tmp_channel "%s\n" s
;;
(*
   \begin{center}
   type: [val prout_newline : unit -> unit]
   \end{center}
*)
let prout_newline () =
  Printf.fprintf !tmp_channel "\n"
;;
(*
   \begin{center}
   type: [val prout_string : string -> unit]
   \end{center}
*)
let prout_string s =
  Printf.fprintf !tmp_channel "%s" s
;;
(*
   \begin{center}
   type: [val prout_int : int -> unit]
   \end{center}
*)
let prout_int i =
  Printf.fprintf !tmp_channel "%d" i
;;


(*
  \section{Miscellaneous}
*)

(*
   The following counter corresponds to the currently parsed HLPSL role.
   We use it to define the scopes of the variables.
   \begin{center}
   type: [val current_role : int ref]
   \end{center}
*)
let current_role = ref 0;;

(*
   initial intruder knowledge
   \begin{center}
   type: [val i_knowledge : Types.if_value_type list ref]
   \end{center}
*)
let i_knowledge = ref ([If_const(global_i)] : if_value_type list);;

(*
   Predicates to be added in the initial section:
   \begin{center}
   type: [val init_predicates : (Types.if_value_type * Types.if_value_type list) list ref]
   \end{center}
*)
let init_predicates = ref ([] : (if_value_type * if_value_type list) list);;

(*
   The programmar may define different depths of debugging.
  \begin{center}
  type: [val debug_level : int ref]
  \end{center}
*)
let debug_level = ref 0;;


(*
  The [sequential_transition_rules] list is a list of transitions 
  implementing the sequential composition.
  \begin{center}
  type: [val sequential_transition_rules : (Types.if_state * Types.if_state) list ref]
  \end{center}
*)
let sequential_transition_rules = ref ([] : (if_state * if_state) list);;
