
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2008 Yann Glouche and Thomas Genet

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.3. Last modification 13/03/08
*)
open Types;;
open Globals;;
open Utils;;
open Error_handler;;
open Interface;;

exception Error_parameters of string;;
exception Undefined_role  of string;;
exception No_function of string;;
exception Invalid_transition;;
exception Invalid_build of string;;
exception Error_start;;
exception Undefined_var of string;;
exception Undefined_const of string;;
exception Error;;

(*
   [compatible_types] checks the compatibility of a type wrt. a second one.
   The first one may be the formal type, and the second one the effective one.
   \Begin{center}
   type: [val compatible_types : term_type * term_type -> bool]
   \end{center}
*)
let rec compatible_types = function
    (Base(Channel_ota),Base(Channel))
  | (Base(Channel),Base(Channel_ota))
  | (Base(Channel_dy),Base(Channel))
  | (Base(Channel),Base(Channel_dy))
  | (Set([]),Set(_))
  | (Set([]),Function(_))
  | (_,Base(Message))
  | (Base(Hash_func),Base(_)) -> true
  | (Set([Pair(t1,Pair(t2,t3))]),Function(s1,s2::ls)) ->
      (compatible_types (t1,s2))
      && (compatible_types (Set([Pair(t1,t3)]),Function(s1,ls)))
  | (Set([Pair(t1,t2)]),Function(s1,[s2])) ->
      (compatible_types (t1,s2))
      && (compatible_types (t2,s1))
  | (x,y) -> x=y
;;
let always_compatible_types = function
  | Base(Message) -> true
  | _ -> false
;;

(*
   \begin{center}
   type:
   [val list_of_error :
   (int * string * string * string * string * string) list ref]
   \end{center}
*)
let list_of_errors = ref [];;

(*
   Utility function transforming embedded HLPSL pairs into a list.
   \begin{center}
   type:
   [val pair_to_list : 'a Types.term_structure -> 'a Types.term_structure list]
   \end{center}
*)
let rec pair_to_list = function
    Pair(y,z) -> y::(pair_to_list z)
  | x -> [x]
;;

(*
   Errors are added to the list errors by [add_sem_error].
   \begin{center}
   type:
   [val add_sem_error : int * string * string * string * string * string -> unit]
   \end{center}
*)
let add_sem_error err =
  list_of_errors := !list_of_errors@[err]
;;


(*
  The [dummy_val_of_type] function associates a constant to a given [term_type] term.
  In this way, only the constants effectively used will be registered and eventually
  printed in the [type] section of the IF file.
   \begin{center}
   type: [val dummy_val_of_type :
          Types.basic_type Types.term_structure ->
          Types.atoms_type Types.term_structure]
   \end{center}
*)
let rec dummy_val_of_type = function
    Base(t) ->
      (try
	(let dummy_value = List.assoc t dummy_values_string_list in
	let dummy_value_id = string_id#register_name dummy_value 0 in
	if (type_table#register_type dummy_value_id (Base(t)) <> (Base(t))) then
	  add_sem_error
	    (20, "constant "^dummy_value^" declared with incorrect type (reserved)", "", "","","");
	Base(Const(dummy_value_id)))
      with Not_found ->
	(let dummy_value_id = string_id#register_name "dummy_msg" 0
	in
	if (type_table#register_type dummy_value_id (Base(Message)) <> (Base(Message))) then
	  add_sem_error
	    (20, "constant dummy_msg declared with incorrect type (reserved)", "", "","","");
	Base(Const(dummy_value_id))))
  | Inv(Base(Public_key)) ->
      Inv(dummy_val_of_type (Base(Public_key)))
  | Pair(t1,t2) ->
      Pair(dummy_val_of_type t1, dummy_val_of_type t2)
  | Scrypt(t1,t2) ->
      Scrypt(dummy_val_of_type t1, dummy_val_of_type t2)
  | Crypt(t1,t2) ->
      Crypt(dummy_val_of_type t1, dummy_val_of_type t2)
  | Function(_) as t -> 
	(let dummy_value_id = string_id#register_name "dummy_func" 0
	in
	if (type_table#register_type dummy_value_id t <> t) then
	  add_sem_error
	    (20, "constant dummy_func declared with incorrect type (reserved)", "", "","","");
	Base(Const(dummy_value_id)))
  | Set(_) -> 
	(let dummy_value_id = string_id#register_name "dummy_set" 0
	in
	if (type_table#register_type dummy_value_id (Set([Base(Message)])) <> (Set([Base(Message)]))) then
	  add_sem_error
	    (20, "constant dummy_set declared with incorrect type (reserved)", "", "","","");
	Base(Const(dummy_value_id)))
(*  | Lst(_) as t -> 
	(let dummy_value_id = string_id#register_name "dummy_list" 0
	in
	if (type_table#register_type dummy_value_id (Lst([])) <> (Lst([]))) then
	  add_sem_error
	    (20, "constant dummy_list declared with incorrect type (reserved)", "", "","","");
	Base(Const(dummy_value_id)))*)
  | _ -> 
	(let dummy_value_id = string_id#register_name "dummy_msg" 0
	in
	if (type_table#register_type dummy_value_id (Base(Message)) <> (Base(Message))) then
	  add_sem_error
	    (20, "constant dummy_msg declared with incorrect type (reserved)", "", "","","");
	Base(Const(dummy_value_id)))
;;


(*
   We define a function to check that an element of a list/set has the
   same type as the list/set.
   We use this function mainly for checking terms of the form: cons(e,L).
   \begin{center}
   type:
   [val compare_types :
   Types.atoms_type Types.term_structure -> Types.term_type -> unit]
   \end{center}
*)
let rec compare_types term term_type =
  match term,term_type with
    (Base(Var(n)),_)
  | (Base(Const(n)),_)
  | (Prime(Var(n)),_ )
  | (Prime(Const(n)),_) ->
      (try
	let type_elet = Globals.type_table#get_type n in
	if not (compatible_types (type_elet,term_type)) then
	  raise (Invalid_build "1")
      with Not_found -> ignore (Globals.type_table#register_type n term_type));
  | (Inv(a1),Inv(a2)) ->
      compare_types a1 a2
  | (Crypt(a1,b1),Crypt(a2,b2))
  | (Scrypt(a1,b1),Scrypt(a2,b2))
  | (Cons(a1,b1),Cons(a2,b2))
  | (Pair(a1,b1),Pair(a2,b2)) ->
      compare_types a1 a2;
      compare_types b1 b2	 
  | (Function(a1,[t1]),Hash(a2)) ->
      compare_types a1 (Base(Hash_func));
      compare_types t1 a2	 
  | (Function(a1,l1),Function(a2,l2)) ->
      compare_types a1 a2;
      List.iter2 compare_types l1 l2	 
  | (Function(a1,_),Base(a2)) ->
      compare_types a1 (Base(a2));
(*  | (Lst([]),Lst(_))*)
  | (Set([]),Set(_)) -> ()
(*  | (Lst(l1),Lst(a2::_))*)
  | (Set(l1),Set([a2])) ->
      List.iter (fun a1 -> compare_types a1 a2) l1;
  | (Set([]),Function(_)) ->
      ()
  | (Set(la1),Function(b2,l2)) ->
      let rec check_each_arg = function
	  (Pair(a1,la1),p2::lp2) ->
	    compare_types a1 p2;
	    check_each_arg (la1,lp2)
	| (a1,[]) ->
	    compare_types a1 b2
	| _ -> invalid_arg "compare_types: incorrect number of arguments"
      in
      List.iter (fun a1 -> check_each_arg (a1,l2)) la1
  | _ ->
      if (not (always_compatible_types term_type)) then
	raise (Invalid_build "1")
;;


(*
   We check that the player of a role is an agent declared as argument
   of the role.
   \begin{center}
   type:
   [val check_player_of_role :
   'a -> int -> hlpsl_variable list -> unit]
   \end{center}
*)
let check_player_of_role player role args =
  if not (List.exists
	    (fun var ->
	      (var#get_id = player)
		&& (var#get_type = Base(Agent)))
	    args)
  then
    add_sem_error(17,"",Globals.string_id#get_name role,"","","")
;;

(* 
   We check the validity of a role call arguments. It is a part of the
   type checking.  The following function retreives the list of role
   parameters and the list of arguments.  Then, it compares the
   corresponding types. If there is a mismatch, we raise an exception
   which indicates the incorrect arguments.
   \begin{center}
   type: [val check_args : call_type -> specification_class -> unit]
   \end{center}
*)
let check_args call specif =
  let list_role =specif#get_roles in

  (* looking for the called role *)
  let r =
    let rec find_role = function
	[] ->
	  raise (Undefined_role (Globals.string_id#get_name(call#get_role)))
      | head::rest ->
	  (match head with
	    Basic(t) ->
	      if t#get_name =call#get_role then head
	      else find_role rest
	  | Composition(t) ->
	      if t#get_name =call#get_role then head
	      else find_role rest)
    in
    find_role list_role  
  in

  let list_param =
    match r with
      Basic(basic_role) -> basic_role#get_param
    | Composition(composition_role) -> composition_role#get_param
  in

  let list_args = call#get_args in

  (* Here is the argument type checking sub-function *) 
  let rec is_type_correct arg param_type =
(*prerr_endline ("is_type_correct:   param_type: "^(atoms_type_to_string false param_type)^"  arg: "^(term_to_string arg)); *TEST*)
    (always_compatible_types param_type)
  ||
    match arg with
      Base(Var(id))
    | Base(Const(id))
    | Prime(Var(id))->
        (try
	  let arg_type = Globals.type_table#get_type id in	
(*prerr_endline ("  arg type: "^(atoms_type_to_string false arg_type)); *TEST*)
	  compatible_types (arg_type,param_type)
	with Not_found -> (* type inference for constants*)
	  ignore (Globals.type_table#register_type id param_type); 
	  true)

    | Function(term,la) ->
(*prerr_endline ("Arg is Function: "^(term_to_string (Function(term,la)))^"  type: "^(atoms_type_to_string false param_type)); *TEST*)
	(* checking that "term" corresponds to an HLPSL function identifier *)
	let term_type =
	  (match term with 
       	    Base(Var(n))
	  | Base(Const(n))
	  | Prime(Var(n)) ->
	      Globals.type_table#get_type n
	  | _ -> raise (No_function ""))
 	in
(*prerr_endline ("   Term_type: "^(atoms_type_to_string false term_type)); *TEST*)
      	(match term_type with
    	  Base(Hash_func) ->
(*	    compatible_types (term_type,param_type) *)
	    (try (compare_types arg param_type; true)
		with _ -> false)
	| Function(r,lp) ->
(*prerr_endline ("   r: "^(atoms_type_to_string false r));
prerr_endline ("  lp: "^(atoms_type_to_string false (List.hd lp))); *TEST*)
	    (compatible_types (r,param_type))
	      &&
	    (* checking that the list of arguments has a valid type*)
	    (let rec check_each_arg = function
		(a1::la1,p2::lp2) ->
		  (is_type_correct a1 p2)
		    && (check_each_arg (la1,lp2))
	      | ([],[]) -> true
	      | _ -> invalid_arg "check_args: incorrect number of arguments"
	    in
	    check_each_arg (la,lp))
        | _ -> false)

    | Set(la) ->
(*prerr_endline ("Arg is Set: "^(term_to_string (Set(la)))); *TEST*)
 	(match param_type with
    	  Function(u,lp) ->
(*prerr_endline ("param_type is Function: "^(atoms_type_to_string false param_type)); *TEST*)
	    (* checking that the list of arguments has a valid type*)
	    let rec check_each_arg = function
		(Pair(a1,la1),p2::lp2) ->
		  (is_type_correct a1 p2)
		    && (check_each_arg (la1,lp2))
	      | (a1,[]) ->
		  (is_type_correct a1 u)
	      | _ -> invalid_arg "check_args: incorrect number of arguments"
	    in
	    List.fold_left
	      (fun b a ->
		(check_each_arg (a,lp)) && b)
	      true la

	  (* Normalizing sets of arguments *)
 	| Set(lp) ->
(*prerr_endline "param_type is Set"; *TEST*)
	    List.fold_left2
	      (fun b a p ->
		(is_type_correct a p) && b)
	      true la lp
	| _ ->
	    false)

    | Cons(e,s)
    | Delete(e,s) ->
	(match param_type with
	  Set([]) -> true
	| Set(t::_) ->
 	    (is_type_correct s param_type)
	      && (is_type_correct e t)
 	| _ -> false)

    | _ ->
(*prerr_endline "Arg is Inv, Pair, Scrypt or Crypt"; *TEST*)
	(match (arg,param_type) with 
	  Inv(t1),Inv(t2) ->
	    is_type_correct t1 t2
	| Pair(t1,t2),Pair(t3,t4) -> 
	    (is_type_correct t1 t3) && (is_type_correct t2 t4)
	| Scrypt(t1,t2),Scrypt(t3,t4) -> 
	    (is_type_correct t1 t3) && (is_type_correct t2 t4)
	| Crypt(t1,t2),Crypt(t3,t4) -> 
	    (is_type_correct t1 t3) && (is_type_correct t2 t4)
	| _,_ -> false)
  in
  (* main call *)
  try
    let n = ref 0 in
(*prerr_endline ("list_param: "^(string_of_int (List.length list_param)));
prerr_endline ("list_args: "^(string_of_int (List.length list_args))); *TEST*)
    List.iter2
      (fun par arg ->
        n := !n + 1;
(*prerr_endline ("List.iter2: n="^(string_of_int !n)); *TEST*)
        if not (is_type_correct arg (par#get_type)) then
  	  raise (Error_parameters (string_of_int !n)))
      list_param list_args;
(*prerr_endline "End of List.iter2" *TEST*)
  with Invalid_argument _ -> raise (Error_parameters ("-1"))
;;

(*
   We check that at least one event is provided in the LHS of HLSPL
   transitions.
   \begin{center}
   type: [val check_event_in_lhs : string -> hlpsl_transition -> unit]
   \end{center}
*)
let check_event_in_lhs msg hlpsl_reaction =
  let rec check = function
      [] ->
	Error_handler.displayWarning 8 msg
    | Event(_,_)::_ -> ()
    | _::rest -> check rest 
  in
  check (hlpsl_reaction#get_lhs)
;;
let check_no_event_in_lhs msg hlpsl_action =
  let rec check = function
      [] -> ()
(*
    | Event(Const(n),[])::rest ->
	if (n = !Globals.global_start) then
	  check rest
	else
	Error_handler.displayWarning 9 msg
*)
    | Event(_,_)::_ ->
	Error_handler.displayWarning 9 msg
    | _::rest -> check rest
  in
  check (hlpsl_action#get_lhs)
;;


(*
   [compare_terms] has the same use as compare, except that it is
   specific to declarations such as: cons(e,[e1,e2,e3..]).
   \begin{center}
   type:
   [val compare_terms :
   Types.atoms_type Types.term_structure ->
   Types.atoms_type Types.term_structure -> unit]
   \end{center}
*)
let compare_terms handle_error =
  let rec compare_terms_rec ter1 ter2 =
    match (ter1,ter2) with
      (_,Base(Var(n2)))
    | (_,Base(Const(n2))) ->
	handle_error
	  (fun _ ->
	    try compare_types ter1 (Globals.type_table#get_type n2)
	    with
 	      Not_found -> raise (Invalid_build "4")
	    | Invalid_argument _ -> raise (Invalid_build "2"))
	  ()
    | (Inv(a1),Inv(a2)) ->
	compare_terms_rec a1 a2
    | (Crypt(a1,b1),Crypt(a2,b2))
    | (Scrypt(a1,b1),Scrypt(a2,b2))
    | (Cons(a1,b1),Cons(a2,b2))
    | (Pair(a1,b1),Pair(a2,b2)) ->
	compare_terms_rec a1 a2;
	compare_terms_rec b1 b2	
    | (Function(a1,l1),Function(a2,l2)) ->
	compare_terms_rec a1 a2;
	List.iter2 compare_terms_rec l1 l2
(*    | (Lst(l1),Lst(l2))*)
    | (Set(l1),Set(l2)) ->
	List.iter2 compare_terms_rec l1 l2
    | _ -> handle_error (fun _ -> raise (Invalid_build "1")) ()
  in
  compare_terms_rec
;;

(*
   We check "cons" terms, corresponding to the construction of a
   list/set by adding an element.
   [check_cons] checks that all the elements have the same type (or a compatible one) as the list/set.
   Exception Invalid\_build is raised if there is a type mismatch.
   \begin{center}
   type:
   [val check_cons : Types.atoms_type Types.term_structure -> unit]
   \end{center}
*)
let check_cons handle_error =
  let rec check_cons_rec = function
      Cons(t1,t2)
    | Delete(t1,t2) ->
	(match t2 with 
	  Base(Var(n))
	| Prime(Var(n)) ->
	    handle_error
	      (fun _ ->
		(try
		  (match (Globals.type_table#get_type n) with
(*		    Lst([])*)
		    Set([]) ->
		      ()
(*		  | Lst(h::_)*)
		  | Set(h::_) ->
		      handle_error
			(fun _ ->
			  try (compare_types t1 h)
			  with Invalid_argument _ -> raise (Invalid_build "2"))
			()
		  | _ -> 
		      handle_error (fun _ -> raise (Invalid_build "3")) ())
		with Not_found ->
		  raise (Undefined_var (Globals.string_id#get_name n))))
	      ()
	| Cons(_)
	| Delete(_) -> check_cons_rec t2 (*test_type t2*)
(*	| Lst(l)*)
	| Set(l) -> List.iter (compare_terms handle_error t1) l
	| _ -> 
	    handle_error (fun _ -> raise (Invalid_build "3")) ())
    | _ -> ()
  in
  check_cons_rec
;;

(*
   We check that the "start" keyword is used as a parameter of a
   channel event.  If the event has no specified type (mainly when is
   a constant), it will be registered in the types table as a channel.
   \begin{center}
   type:
   [val check_start : Types.predicate -> unit]
   \end{center}
*)				
let check_start handle_error = function
    Event(t1, Base(Const(n))::_) ->
      if (Globals.string_id#get_name n) = "start" then
	(match t1 with
  	  Var(n)
 	| Const(n) ->
  	    (try
  	      (match (Globals.type_table#get_type n) with
  		Base(x)
	      | Prime(x) ->
  		  (match x with
 		    Channel
		  | Channel_dy
		  | Channel_ota -> ()
 		  | _ -> handle_error (fun _ -> raise Error_start) ())
  	      | _ -> ())
  	    with Not_found ->
	      ignore (Globals.type_table#register_type n (Base(Channel)))))
  | _ -> ()
;;

(*
   The function [check_var] verifies recursively that the variables of
   a term have been declared.  It is based on types tables because
   undeclared variables have no entries there.
   \begin{center}
   type:
   [val check_var : Types.atoms_type Types.term_structure -> unit]
   \end{center}
*)
let check_var handle_error =
  let rec check_var_rec = function
      Base(Var(n))
    | Prime(Var(n)) ->
	handle_error
	  (fun n ->
	    try ignore (Globals.type_table#get_type n)
	    with Not_found ->
	      raise (Undefined_var (Globals.string_id#get_name n)))
	  n
    | Base(Const(n)) ->
	handle_error
	  (fun n ->
	    try ignore (Globals.type_table#get_type n)
	    with Not_found ->
	      raise (Undefined_const (Globals.string_id#get_name n)))
	  n
    | Prime(_) -> ()
    | Pair(t1,t2)
    | Scrypt(t1,t2)
    | Crypt(t1,t2)
    | Cons(t1,t2)
    | Delete(t1,t2) -> check_var_rec t1; check_var_rec t2
    | Function(t1,t2) -> check_var_rec t1; List.iter check_var_rec t2
    | Inv(t) -> check_var_rec t
(*    | Lst(l)*)
    | Set(l) -> List.iter check_var_rec l
    | Hash(_) ->
	prerr_endline "Semantic_check.check_var: Impossible error: invalid hash type/term"; 
	raise Not_found
  in
  check_var_rec
;;

(*
   We check that LHS of the assignment operator "=" is a variable.
   \fix{check if we need it to be primed}.
   \begin{center}
   type:
   [val check_equal : Types.predicate -> unit]
   \end{center}
*)
let check_equal handle_error =
  let rec check_equal_rec = function
      Not(p) ->
	check_equal_rec p
    | Equal(Base(Var(_)) as t1, t2)
    | Equal(Prime(Var(_)) as t1, t2) ->
	compare_terms handle_error t2 t1
    | Equal(_) ->
	raise Error
    | _ -> ()
  in
  check_equal_rec
;;

(*
   We check that variables declared in local section appended to role
   parameters are not duplicated.
   We prevent also the redeclaration of owned variables.
   \begin{center}
   type:
   [val check_duplication : generic_hlpsl_role -> unit]
   \end{center}
*)
let check_duplication role =
  let role_name = Globals.string_id#get_name role#get_name in
  (* recursive check of a list of vars *)
  let rec check_duplicates list_string = function
      [] -> ()
    | head::tail ->
 	if List.exists (fun x -> x#get_id=head#get_id) tail then
	  add_sem_error
	    (13,(Globals.string_id#get_name head#get_id),
	     role_name,"","",(list_string^" list"));
	check_duplicates list_string tail
  in
  check_duplicates "parameters or local" (role#get_param@role#get_local);
  check_duplicates "owned" role#get_owned
;;

(*
   In order to check predicates variables, we define a recursive
   function [check_predicate_vars].
   \begin{center}
   type:
   [val check_predicate_vars : ]
   \end{center}
*)
let check_predicate_vars handle_error =
  let rec check_predicate_vars_rec = function
      Not(p) ->
	check_predicate_vars_rec p
    | Equal(t1,t2)
    | Leq(t1,t2)
    | In(t1,t2) ->
	check_var handle_error t1;
	check_var handle_error t2
    | Event(Const(i),lt) ->
	List.iter (check_var handle_error) lt
    | Event(t,lt) ->
	check_var handle_error (Base t);
	List.iter (check_var handle_error) lt
    | _ -> ()
  in
  check_predicate_vars_rec
;;

(*
   The next function, [check_goals] tries to detect some errors in the
   goals specification.
   \begin{center}
   type:
   [val check_goals : Types.goals_type list -> unit]
   \end{center}
*)
let check_goals =
  let t = Globals.string_id#get_tab_string_to_id in
  let is_typed n =
    let type_found = ref false in
    Hashtbl.iter
      (fun (st,_) idst ->
	if (n = st) then
	  try
 	    ignore (Globals.type_table#get_type idst);
	    type_found := true
	  with Not_found -> ())
      t;
    !type_found
  in
  List.iter
    (function
	LTL_goal(_,lvc) ->
	  List.iter
	    (function
		Base(Var(n)) ->
		  let nst = Globals.string_id#get_name n in
		  if not (is_typed nst) then
		    add_sem_error 
		      (11, nst, "","","","")
	      | Base(Const(n)) ->
		  let nst = Globals.string_id#get_name n in
		  if not (is_typed nst) then
		    Error_handler.displayWarning
		      7 (nst^" in LTL goal")
	      | _ -> ())
	    lvc

      | Secrecy(lc) ->
	  List.iter
	    (fun n ->
	      let nst = Globals.string_id#get_name n in
	      if not (is_typed nst) then
		Error_handler.displayWarning
		  7 (nst^" in Secrecy goal"))
	    lc

      | Authentication(lc) ->
	  List.iter
	    (fun n ->
	      let nst = Globals.string_id#get_name n in
	      if not (is_typed nst) then
		Error_handler.displayWarning
		  7 (nst^" in Authentication goal"))
	    lc

      | Weak_authentication(lc) ->
	  List.iter
	    (fun n ->
	      let nst = Globals.string_id#get_name n in
	      if not (is_typed nst) then
		Error_handler.displayWarning
		  7 (nst^" in Weak authentication goal"))
	    lc

      | Auth(r1,r2,_) ->
	  (try ignore (Globals.string_id#get_name (find_player_in_role r1))
	  with Not_found ->
	    add_sem_error(20,"in authentication goal, invalid role name "^(Globals.string_id#get_name r1),"","","",""));
	  (try ignore (Globals.string_id#get_name (find_player_in_role r2))
	  with Not_found ->
	    add_sem_error(20,"in authentication goal, invalid role name "^(Globals.string_id#get_name r2),"","","",""))

      | Weak_auth(r1,r2,_) ->
	  (try ignore (Globals.string_id#get_name (find_player_in_role r1))
	  with Not_found ->
	    add_sem_error(20,"in weak authentication goal, invalid role name "^(Globals.string_id#get_name r1),"","","",""));
	  (try ignore (Globals.string_id#get_name (find_player_in_role r2))
	  with Not_found ->
	    add_sem_error(20,"in weak authentication goal, invalid role name "^(Globals.string_id#get_name r2),"","","","")))
;; 


(*
   The main function in this module is [check_all]. It performs calls
   to semantic check functions and handles the raised exceptions.
   \begin{center}
   type:
   [val check_all : specification_class -> unit]
   \end{center}
*)
let check_all specif =
  let generic_role_check role role_name =
    (*cheking duplicates*)
    check_duplication role; 

    (* iterative check of initial state predicates *)
    List.iter
      (fun init_predicates ->
	(* Error handling *)
	let handle_error f t =
      	  try f t
      	  with
	    Undefined_var s -> 
	      add_sem_error 
		(11,s,role_name,"","init state","")
      	  | Undefined_const s -> 
	      Error_handler.displayWarning
		7 (s^" in init state of role "^role_name)
	in
      	(* recursive check of variables declaration *)
	check_predicate_vars handle_error init_predicates;
	(* Error handling *)
	let handle_error f t =
      	  try f t
      	  with
	    Error ->
	      add_sem_error
		(12,"",role_name,"","init state","")
	  | Invalid_build s ->
	      add_sem_error
		(5,s^"i",role_name,"","","")
	in
	check_equal handle_error init_predicates)
      role#get_init;

    (* iterative check of accept state predicates *)
    List.iter
      (fun pr ->
	let handle_error f t =
      	  try f t
      	  with
	    Undefined_var s -> 
	      add_sem_error
		(11,s,role_name,"","accept state","")
      	  | Undefined_const s -> 
	      Error_handler.displayWarning
		7 (s^" in accept state of role "^role_name)
	in
	check_predicate_vars handle_error pr)
      role#get_accept;

    (* iterative check of knowledge declarations * Removed on April 6, 2005
    List.iter
      (fun (a,lt) ->
      	* recursive check of variables declaration *
	let handle_error f t =
      	  try f t
      	  with
	      Undefined_var s -> 
		add_sem_error 
		  (11,s,role_name,"","knowledge declaration","")
      	    | Undefined_const s -> 
		Error_handler.displayWarning
		  7 (s^" in knowledge declaration of role "^role_name)
      	    | Error -> 
		add_sem_error
		  (12,"",role_name,"","knowledge declaration","")
	in
      	List.iter (check_var handle_error) lt)
      role#get_knowledge*)
    (* check of variables declaration *)
    let handle_error f t =
      try f t
      with
	Undefined_var s -> 
	  add_sem_error 
	    (11,s,role_name,"","intruder knowledge declaration","")
      | Undefined_const s -> 
	  Error_handler.displayWarning
	    7 (s^" in intruder knowledge declaration of role "^role_name)
      | Error -> 
	  add_sem_error
	    (12,"",role_name,"","intruder knowledge declaration","")
    in
    List.iter (check_var handle_error) role#get_knowledge
  in

  List.iter
    (function
      	Basic(basic_role) ->
	  let name = Globals.string_id#get_name basic_role#get_name in
	  (* HLPSL role generic checks *)
	  generic_role_check basic_role name;

	  (* Basic roles specific checks *)
	  let actions = basic_role#get_actions in
	  let check_transition =
	    (fun (is_action,tr) ->
              let name_tr = Globals.string_id#get_name tr#get_label in
	      (* Error handling *)
	      let handle_error f t =
		try
		  f t
		with
	    	  Undefined_var s -> add_sem_error(11,s,name,name_tr,"","")
      		| Undefined_const s -> 
		    Error_handler.displayWarning
		      7 (s^" in transition "^name_tr^" of role "^name)
		| Invalid_transition -> add_sem_error(4,"",name,name_tr,"","")
		| Invalid_build s -> add_sem_error(5,s^"t",name,name_tr,"","")
		| Error_start -> add_sem_error(10,"",name,name_tr,"","")
	      in
	      (* recursive check of variables and constants *)
	      let rec check_atoms = function
		  Not(p) ->
		    check_atoms p
       		| New(t) ->
		    check_var handle_error (Base(Var(t)))
       		| Equal(t1,t2) ->
		    check_var handle_error t1;
		    check_var handle_error t2;
		    check_cons handle_error t1;
		    check_cons handle_error t2
(***TO BE ADDED: but with more flexibility...
		    compare_terms handle_error t2 t1
 ***)
		| Leq(t1,t2)
       		| In(t1,t2) ->
		    check_var handle_error t1;
		    check_var handle_error t2;
		    check_cons handle_error t1;
		    check_cons handle_error t2
                | Event(Const(i),lt) ->
		    (* constant predicates do not have to be declared *)
		    List.iter (check_var handle_error) lt;
		    check_cons handle_error (Base(Const(i)));
		    List.iter (check_cons handle_error) lt
                | Event(t,lt) ->
		    check_var handle_error (Base t);
		    List.iter (check_var handle_error) lt;
		    check_cons handle_error (Base t);
		    List.iter (check_cons handle_error) lt
              in
	    	  (* iterative check of hlpsl transition states *)
              List.iter
		(fun state ->
		   check_atoms state;
		   check_start handle_error state)
		(tr#get_lhs@tr#get_rhs);
	      if is_action then
		check_no_event_in_lhs ("role "^name^", transition "^name_tr) tr
	      else
		check_event_in_lhs ("role "^name^", transition "^name_tr) tr)
	  in
	  List.iter check_transition actions;

      | Composition(composition_role) ->
	    let name = Globals.string_id#get_name composition_role#get_name in

	    let comp = composition_role#get_composition in
	      (* handling calls over declarations: variables type inference *)
	      
	    let handle_multi_dec multi_call =
 	      let term = multi_call#get_term in
 	      let var_list = pair_to_list term in
	      let type_list =
		match (type_table#get_type (multi_call#get_set)) with
	       	    Function(in_type,[out_type]) ->
		      [in_type]@(pair_to_list out_type)
(*		  | Lst([elem_type])*)
		  | Set([elem_type]) ->
		      pair_to_list elem_type
		  | _ -> [Base(Message)]
	      in
	      let handle_error f t =
		try f t
		with
		  Undefined_var s -> 
		    add_sem_error 
		      (11,s,name,"","multi parallel composition","")
		| Undefined_const s -> 
		    Error_handler.displayWarning
		      7 (s^" in multi parallel composition of role "^name)
		| Error -> 
		    add_sem_error
		      (12,"",name,"","multi parallel composition","")
	      in
	      List.iter (check_var handle_error) var_list;
	      List.iter2
		(fun var tp ->
		  ignore
		    (Globals.type_table#register_type
		       (match var with
			 Base(Var i) -> i
		       | _ -> 
			   Error_handler.displayWarning
			     2 "handle_multi_dec";	
			   0)
		       tp))
		var_list type_list;
	      check_cons
		(fun f t ->
		  try f t
		  with 
		    Invalid_argument _ -> 
		      add_sem_error (15,"",name,"","",""))
		term;
	    in
	      
	    let rec scan_composition = function
    		Call(c) ->
    	      	  let name_cal = Globals.string_id#get_name c#get_role in
		  let handle_error f t =
		    try
		      f t
		    with
		      Undefined_var s -> 
			add_sem_error(11,s,name,"","",name_cal)
      		    | Undefined_const s -> 
			Error_handler.displayWarning
			  7 (s^" in Call "^name_cal^" of role "^name)
	      	    | Undefined_role s -> 
			add_sem_error(0,"",s,"","","")
	      	    | No_function s -> 
			add_sem_error(1,s,name,"","",name_cal)
	      	    | Invalid_build s -> 
			add_sem_error(5,s^"c",name,"","",name_cal)
	      	    | Error_parameters s -> 
			add_sem_error(2,s,name,"","",name_cal)
		  in
    		  List.iter
		    (fun term ->
		      check_var handle_error term;
		      check_cons handle_error term)
		    c#get_args;
		  handle_error (check_args c) specif
		    
	      | MultiPar(multi_call,composition) ->
(*	      | MultiSeq(multi_call,composition) *)
		  handle_multi_dec multi_call;
	      	  scan_composition composition
		    
	      | Sequential(t1,t2) 
	      | Parallel(t1,t2) ->
		  scan_composition t1;
		  scan_composition t2
		    
	      | Empty ->
		  ()
	    in 

	    try 
	      generic_role_check composition_role name;
	      scan_composition comp 
	    with
	      Undefined_var s -> add_sem_error(11,s,name,"","","")
	    | Invalid_build s -> add_sem_error(5,s,name,"","",""))
    specif#get_roles;
  
  (* Verification of multi-declarations of variables: they have to be of the
     same type, even if they are declared in different roles. *)

  let t = Globals.string_id#get_tab_string_to_id in
  Hashtbl.iter
    (fun (st,idrole1) idst ->
      try
 	(let type_idst = Globals.type_table#get_type idst in
      	 Hashtbl.iter
    	   (fun (st2,idrole2) idst2 ->
	     if (st = st2) && (idrole1 < idrole2) then
	       try
		 (let type_idst2 = Globals.type_table#get_type idst2 in
      	      	  if not (compatible_types (type_idst,type_idst2)) then
		    add_sem_error(18,st, (try Globals.string_id#get_name idrole1 with _ -> ""),
				  (try Globals.string_id#get_name idrole2 with _ -> ""),"",""))
	       with Not_found -> ())
	   t)
      with Not_found -> ())
    t;

  check_goals specif#get_goals;

  if (!list_of_errors <> []) then
    (List.iter Error_handler.handleSemanticError !list_of_errors;
     if !debug_level >= 5 then
       Utils_debug.dump();
     exit(-1))
;;
